#include "../include/disk.h"
#include "../include/tga.h"
#include <glib.h>
#include "editortex.h"
#include "config.h"


gint EditorTexLoad(editor_tex_struct *t, const gchar *path);
editor_tex_struct *EditorTexNew(void);
editor_tex_struct *EditorTexNewLoad(
        const gchar *path,              /* Loaded from this file. */
        const gchar *name,
        const gchar *filename,
        gfloat priority
);
void EditorTexDelete(editor_tex_struct *t);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)     (((x) < 0) ? ((x) * -1) : (x))
#define IS_STRING_EMPTY(s)      (((s) != NULL) ? (*(s) == '\0') : TRUE)

#define RADTODEG(r)     ((r) * 180 / PI)
#define DEGTORAD(d)     ((d) * PI / 180)


/*
 *	Loads the new texture specified by path to the texture structure
 *	t, any existing texture on t will be unloaded.
 *
 *	Returns non-zero on error.
 */
gint EditorTexLoad(editor_tex_struct *t, const gchar *path)
{
	gchar *dpath;


	if((t == NULL) || (path == NULL))
	    return(-1);

	if(*path == '\0')
	    return(-1);

	/* Complete path as needed. */
	if(ISPATHABSOLUTE(path))
	{
	    dpath = STRDUP(path);
	}
	else
	{
	    dpath = g_strdup_printf(
		"%s%c%s",
		SE_DEF_SAR_DATA_DIR,
		DIR_DELIMINATOR,
		path
	    );
	}

	/* Unload existing texture if any. */
	if(t->data != NULL)
	{
	    g_free(t->data);
	    t->data = NULL;
	}

        /* Load data from file, read as RGBA 32 bits. */
	t->data = (guint8 *)TgaReadFromFileFastRGBA(
	    dpath,
	    &t->width, &t->height,
	    0xff000000	/* Transparent pixel rgba. */
	);
	g_free(dpath);
	dpath = NULL;

	t->bpp = 4;
	t->bpl = t->width * t->bpp;

	return(0);
}

/*
 *	Creates a new texture structure.
 */
editor_tex_struct *EditorTexNew(void)
{
	return(EDITOR_TEX(g_malloc0(
	    sizeof(editor_tex_struct)
	)));
}

/*
 *	Creates a new texture structure with everything loaded.
 */
editor_tex_struct *EditorTexNewLoad(
        const gchar *path,              /* Loaded from this file. */
        const gchar *name,
        const gchar *filename,
        gfloat priority
)
{
	editor_tex_struct *t = EditorTexNew();
	if(t == NULL)
	    return(NULL);

	t->name = STRDUP(name);
	t->filename = STRDUP(filename);
	t->priority = priority;
	t->data = NULL;

	EditorTexLoad(t, path);

	return(t);
}

/*
 *	Deletes the texture.
 */
void EditorTexDelete(editor_tex_struct *t)
{
	if(t == NULL)
	    return;

	g_free(t->name);
	g_free(t->filename);
	g_free(t->data);
	g_free(t);
}
