#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef __MSW__
# include <windows.h>
# include "resource.h"
#else
# include <unistd.h>
#endif

#include <gtk/gtk.h>
#include <gdk/gdkrgb.h>

#include "../include/string.h"
#include "../include/strexp.h"
#include "../include/disk.h"

#include "guiutils.h"
#include "cdialog.h"
#include "fb.h"
#include "csd.h"
#include "pdialog.h"
#include "fprompt.h"

#include "editor.h"
#include "editorobj.h"
#include "editorscnopen.h"
#include "editorview.h"
#include "options.h"
#include "scnedit.h"
#include "config.h"


static void SESignalCB(int s);
static gint SETimeoutCB(gpointer data);

static se_core_struct *SENew(
        gint argc, gchar *argv[],
        gint *err_rtn
);
static void SEDelete(se_core_struct *core_ptr, gboolean save_config);

se_dde_struct *SEDDENew(gpointer core_ptr, gint editor_num);
void SEDDEObjectAdd(se_dde_struct *dde, gpointer obj);
void SEDDEDelete(se_dde_struct *dde);


static se_core_struct *se_core;


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define IS_STRING_EMPTY(s)      (((s) != NULL) ? (*(s) == '\0') : TRUE)

#define RADTODEG(r)     ((r) * 180 / PI)
#define DEGTORAD(d)     ((d) * PI / 180)


/*
 *	UNIX signal callback.
 */
static void SESignalCB(int s)
{
	static int segfault_count = 0;
	se_core_struct *core_ptr = se_core;
	if(core_ptr == NULL)
	    return;

	switch(s)
	{
	  case SIGINT:
	  case SIGTERM:
	    core_ptr->close_all_windows = TRUE;
	    break;
	  case SIGSEGV:
            segfault_count++;   /* Count this segfault. */
            fprintf(
                stderr,
 "%s triggered a segmentation fault (%i times)\n",
                PROG_NAME,
                segfault_count
            );
            /* 3 or more segfaults? */
            if(segfault_count >= 3)
            {
                fprintf(
                    stderr,
 "%s attempting immediate process exit().\n",
                    PROG_NAME
                );
                exit(1);        /* Explicit exit. */
            }
            else
            {
                core_ptr->close_all_windows = TRUE;
            }
            break;
#ifdef SIGHUP
          case SIGHUP:
            break;
#endif
#ifdef SIGPIPE
          case SIGPIPE:
            break;
#endif
	}
}

/*
 *	Timeout callback.
 */
static gint SETimeoutCB(gpointer data)
{
	gboolean all_windows_closed = TRUE;
        editor_struct *e;
	se_core_struct *core_ptr = SE_CORE(data);
	if(core_ptr == NULL)
	    return(FALSE);

	/* Check editor. */
	e = core_ptr->editor;
	if(all_windows_closed && EDITOR_IS_MAPPED(e))
	    all_windows_closed = FALSE;

        /* When all windows are closed, do not call this timeout anymore
         * and break out of the main GTK+ loop.
         */
        if(all_windows_closed || core_ptr->close_all_windows)
        {
            core_ptr->toid = (guint)-1;
            gtk_main_quit();
            return(FALSE);
        }
        else
        {
            return(TRUE);
        }
}

/*
 *	Creates a new core structure.
 */
static se_core_struct *SENew(
	gint argc, gchar *argv[],
	gint *err_rtn
)
{
	gchar *startup_scn_filename = NULL;
	gchar *cfg_file = NULL;
	gchar *sar_dir = NULL;
	gint i;
	const gchar *arg;
	editor_struct *e;
	options_struct *opt;
	se_core_struct *core_ptr;


	*err_rtn = 0;


        /* Handle command line arguments. */
#ifdef __MSW__
	for(i = 0; i < argc; i++)
#else
        for(i = 1; i < argc; i++)
#endif
        {
            arg = argv[i];
            if(arg == NULL)
                continue;

	    /* Configuration file. */
	    if(!g_strcasecmp(arg, "--configuration") ||
	       !g_strcasecmp(arg, "-configuration") ||
	       !g_strcasecmp(arg, "--config") ||
	       !g_strcasecmp(arg, "-config") ||
               !g_strcasecmp(arg, "--f") ||
               !g_strcasecmp(arg, "-f")
	    )
	    {
		i++;
		arg = (i < argc) ? argv[i] : NULL;
		if(arg != NULL)
		{
		    g_free(cfg_file);
		    cfg_file = STRDUP(arg);
		}
		else
		{
		    arg = argv[i - 1];
		    fprintf(
			stderr,
"%s: Requires at least one argument.\n",
			arg
		    );
		}
	    }
            /* Search and Rescue directory. */
            else if(!g_strcasecmp(arg, "--sar-directory") ||
                    !g_strcasecmp(arg, "-sar-directory") ||
                    !g_strcasecmp(arg, "--sar-dir") ||
                    !g_strcasecmp(arg, "-sar-dir") ||
                    !g_strcasecmp(arg, "--d") ||
                    !g_strcasecmp(arg, "-d")
	    )
	    {
		i++;
                arg = (i < argc) ? argv[i] : NULL;
                if(arg != NULL)
                {
		    struct stat stat_buf;

                    g_free(sar_dir);
                    sar_dir = STRDUP(arg);

		    if(stat(sar_dir, &stat_buf))
			fprintf(
			    stderr,
"%s: Warning: Unable to access directory `%s'.\n",
			    argv[i - 1], sar_dir
			);
#ifdef S_ISDIR
		    else if(!S_ISDIR(stat_buf.st_mode))
			fprintf(
                            stderr,
"%s: Warning: `%s' is not a directory.\n",
                            argv[i - 1], sar_dir
                        );
#endif
                }
                else
                {
                    arg = argv[i - 1];
                    fprintf(
                        stderr,
"%s: Requires at least one argument.\n",
                        arg
                    );
                }
	    }
            /* All else assume address? */
            else if((*arg != '+') && (*arg != '-'))
	    {
		g_free(startup_scn_filename);
		startup_scn_filename = g_strdup(arg);
            }
	}

	/* Use default configuration file? */
	if(cfg_file == NULL)
	{
#ifdef __MSW__
	    cfg_file = STRDUP(SE_DEF_INI_FILE);
#else
	    const gchar *home = getenv("HOME");
	    cfg_file = g_strdup_printf(
		"%s/%s",
		(home != NULL) ? home : "",
		SE_DEF_INI_FILE
	    );
#endif
	}

	/* Allocate new core structure. */
	core_ptr = SE_CORE(g_malloc0(sizeof(se_core_struct)));
	if(core_ptr == NULL)
	{
	    *err_rtn = 3;
	    return(NULL);
	}

	/* Get startup directory, this is mainly for Windows since it
	 * is difficult to find the current working directory so for
	 * Window specific code we'll use this directory.
	 */
	if(TRUE)
	{
	    gchar cwd[PATH_MAX];
	    if(getcwd(cwd, PATH_MAX) != NULL)
		cwd[PATH_MAX - 1] = '\0';
	    else
		*cwd = '\0';
	    core_ptr->cwd = STRDUP(cwd);
	}

	core_ptr->close_all_windows = FALSE;
	core_ptr->toid = (guint)-1;
	core_ptr->dde = NULL;

        /* Load cursors. */
        core_ptr->cursor_busy = gdk_cursor_new(GDK_WATCH);
        core_ptr->cursor_text = gdk_cursor_new(GDK_XTERM);
        core_ptr->cursor_translate = gdk_cursor_new(GDK_FLEUR);
        core_ptr->cursor_zoom = gdk_cursor_new(GDK_SIZING);
        core_ptr->cursor_rect_select = gdk_cursor_new(GDK_CROSSHAIR);

	/* Create options structure and load options from file. */
	core_ptr->options = opt = OptionsNew();
	OptionsLoadFromFile(opt, cfg_file);

	/* Overwrite options from values obtained from command line. */
	/* Search and Rescue directory. */
	if(sar_dir != NULL)
	{
	    g_free(opt->sar_dir);
	    opt->sar_dir = STRDUP(sar_dir);
	}

	/* Global tooltips state. */
        GUISetGlobalTipsState(TRUE);

	/* Create a new editor. */
	core_ptr->editor = e = EditorNew(core_ptr);
	EditorMap(e);
	/* Load scenery at startup? */
	if(startup_scn_filename == NULL)
	    startup_scn_filename = STRDUP(core_ptr->options->last_scn_filename);
	if((startup_scn_filename != NULL) ?
	    (*startup_scn_filename != '\0') : FALSE
	)
	{
	    while(gtk_events_pending() > 0)
		gtk_main_iteration();

	    /* Reset editor view position from last position. */
	    e->view_cursor_xw = opt->editor_last_cursor_xw;
	    e->view_cursor_yw = opt->editor_last_cursor_yw;
	    EditorViewSetViewPositionM(
		e, opt->editor_last_view_xm, opt->editor_last_view_ym
	    );
	    EditorViewSetViewZoomM(e, opt->editor_last_view_zoom);
	    EditorViewSetViewGridSpacingM(e, opt->editor_last_view_grid_spacing); 

	    /* Load scenery. */
	    EditorSetBusy(e, TRUE);
	    EditorSceneryLoad(e, startup_scn_filename, TRUE);
	    EditorViewDraw(e);
	    EditorUpdateTitle(e);
	    EditorUpdateMenus(e);
	    EditorSetBusy(e, FALSE);
	}

        /* Set timeout callbacks. */
        core_ptr->toid = gtk_timeout_add(
            100, (GtkFunction)SETimeoutCB, core_ptr
        );

	/* Deallocate locals. */
	g_free(startup_scn_filename);
	g_free(sar_dir);
	g_free(cfg_file);

	return(core_ptr);
}

/*
 *	Deletes the core structure.
 */
static void SEDelete(se_core_struct *core_ptr, gboolean save_config)
{
	GdkCursor **cursor;


	if(core_ptr == NULL)
	    return;

	/* Save configuration? */
	if(save_config && (core_ptr->options != NULL))
	{
	    const options_struct *opt = core_ptr->options;
	    OptionsSaveToFile(opt, opt->cfg_file);
	}

	SEDDEDelete(core_ptr->dde);
	core_ptr->dde = NULL;

	/* Remove timeout callbacks. */
	if(core_ptr->toid != (guint)-1)
	{
	    gtk_timeout_remove(core_ptr->toid);
	    core_ptr->toid = (guint)-1;
	}

	/* Delete editor. */
	EditorDelete(core_ptr->editor);
	core_ptr->editor = NULL;

	/* Delete options. */
	OptionsDelete(core_ptr->options);
	core_ptr->options = NULL;

	/* Destroy cursors. */
#define DO_DESTROY_CURSOR	\
{ if(*cursor != NULL) { gdk_cursor_destroy(*cursor); *cursor = NULL; } }
	cursor = &core_ptr->cursor_busy;
	DO_DESTROY_CURSOR
	cursor = &core_ptr->cursor_text;
	DO_DESTROY_CURSOR
        cursor = &core_ptr->cursor_translate;
        DO_DESTROY_CURSOR
        cursor = &core_ptr->cursor_zoom;
        DO_DESTROY_CURSOR
        cursor = &core_ptr->cursor_rect_select;
        DO_DESTROY_CURSOR
#undef DO_DESTROY_CURSOR

	g_free(core_ptr->cwd);

	/* Deallocate structure itself. */
	g_free(core_ptr);
}


/*
 *	Creates a new DDE structure.
 */
se_dde_struct *SEDDENew(gpointer core_ptr, gint editor_num)
{
	se_dde_struct *dde = SE_DDE(g_malloc0(sizeof(se_dde_struct)));
	if(dde == NULL)
	    return(NULL);

	dde->core_ptr = core_ptr;
	dde->editor_num = editor_num;
	dde->object = NULL;
	dde->total_objects = 0;

	return(dde);
}

/*
 *	Adds object to the dde.
 */
void SEDDEObjectAdd(se_dde_struct *dde, gpointer obj)
{
	gint i;

        if(dde == NULL)
	{
	    EditorObjDelete(obj);
            return;
	}

	i = MAX(dde->total_objects, 0);
	dde->total_objects = i + 1;

	dde->object = (gpointer *)g_realloc(
	    dde->object, dde->total_objects * sizeof(gpointer)
	);
	dde->object[i] = obj;
}

/*
 *	Deletes the DDE structure.
 */
void SEDDEDelete(se_dde_struct *dde)
{
	gint i;

        if(dde == NULL)
            return;

	for(i = 0; i < dde->total_objects; i++)
	    EditorObjDelete(EDITOR_OBJ(dde->object[i]));
	g_free(dde->object);

	g_free(dde);
}

#ifdef __MSW__
int WINAPI WinMain(
	HINSTANCE hInstance,		// Instance
	HINSTANCE hPrevInstance,	// Previous Instance
	LPSTR lpCmdLine,		// Command Line Parameters
	int nCmdShow			// Window Show State
)
#else
int main(int argc, char *argv[])
#endif
{
        gboolean initialized_gtk = FALSE;
	gint i, err_val;
	const gchar *arg;
	se_core_struct *core_ptr;
#ifdef __MSW__
	char **argv;
	int argc;
#endif

	se_core = NULL;
	tzset();

#ifdef __MSW__
	argv = strexp(lpCmdLine, &argc);
	CoInitialize(NULL);
#endif

	/* Handle basic command line arguments. */
#ifdef __MSW__
	for(i = 0; i < argc; i++)
#else
	for(i = 1; i < argc; i++)
#endif
	{
	    arg = argv[i];
	    if(arg == NULL)
		continue;

	    /* Help? */
	    if(!g_strcasecmp(arg, "--help") ||
	       !g_strcasecmp(arg, "-help") ||
	       !g_strcasecmp(arg, "--h") ||
               !g_strcasecmp(arg, "-h") ||
               !g_strcasecmp(arg, "-?")
	    )
	    {
#ifndef __MSW__
		printf(PROG_HELP_MESG);
#endif
		return(0);
	    }
            /* Version? */
            else if(!g_strcasecmp(arg, "--version") ||
                    !g_strcasecmp(arg, "-version")
            )
            {
		printf(
		    "%s Version %s\n%s",
		    PROG_NAME, PROG_VERSION,
		    PROG_COPYRIGHT
		);
                return(0);
            }
	}


	/* Set up UNIX signal callbacks. */
	signal(SIGINT, SESignalCB);
        signal(SIGTERM, SESignalCB);
        signal(SIGSEGV, SESignalCB);
#ifdef SIGHUP
        signal(SIGHUP, SESignalCB);
#endif
#ifdef SIGPIPE
        signal(SIGPIPE, SESignalCB);
#endif

	/* Reset timezone. */
	tzset();


        /* Initialize GTK+ as needed. */
        if(!initialized_gtk)
        {
            if(!gtk_init_check(&argc, &argv))
            {
                fprintf(
                    stderr,
 "This program requires X.\n"
                );
                return(1);
            }
            initialized_gtk = TRUE;
        }

        gdk_rgb_init();

        CDialogInit();
        FileBrowserInit();
	CSDInit();
	PDialogInit();
	FPromptInit();

	/* Create core structure. */
	se_core = core_ptr = SENew(
	    argc, argv, &err_val
	);

	/* If core is successfully created then enter main GTK+ loop. */
	if(core_ptr != NULL)
	    gtk_main();

	/* Delete core structure. */
	se_core = NULL;
	SEDelete(core_ptr, TRUE);
	core_ptr = NULL;

	FPromptShutdown();
	PDialogShutdown();
	CSDShutdown();
        FileBrowserShutdown();
	CDialogShutdown();

#ifdef __MSW__
	CoUninitialize();
	StringFreeArray(argv, argc);
#endif

	return(err_val);
}
